import pytz
from unittest import mock
from datetime import datetime, timedelta
from django.test import TestCase
from allianceauth.eveonline.models import EveCorporationInfo
from eveuniverse.models import EveType, EveMoon, EveGroup, EveCategory, EveSolarSystem, EvePlanet, EveConstellation, EveRegion

from moonstuff.templatetags import filters
from ..models import Resource, Refinery, Extraction


class TestFilters(TestCase):
    def setUp(self):
        # Moon and Resource Data
        EveCategory.objects.create(id=1, published=True)
        EveCategory.objects.create(id=2, published=True)
        EveGroup.objects.create(id=1884, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1920, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1921, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1922, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1923, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1, published=True, eve_category_id=1)
        EveType.objects.create(id=1, published=True, eve_group_id=1884)
        EveType.objects.create(id=2, published=True, eve_group_id=1920)
        EveType.objects.create(id=3, published=True, eve_group_id=1921)
        EveType.objects.create(id=4, published=True, eve_group_id=1922)
        EveType.objects.create(id=5, published=True, eve_group_id=1923)
        EveType.objects.create(id=6, published=True, eve_group_id=1)
        EveRegion.objects.create(id=1)
        EveConstellation.objects.create(id=1, eve_region_id=1)
        EveSolarSystem.objects.create(id=1, security_status=1, eve_constellation_id=1)
        EvePlanet.objects.create(id=1, eve_solar_system_id=1, eve_type_id=6)
        EvePlanet.objects.create(id=2, eve_solar_system_id=1, eve_type_id=6)
        self.moon = EveMoon.objects.create(id=1, eve_planet_id=1)
        self.moon2 = EveMoon.objects.create(id=2, eve_planet_id=2)
        self.moon3 = EveMoon.objects.create(id=3, eve_planet_id=2)

        self.rs1 = Resource.objects.create(
            ore_id=1,
            quantity=0.234445456,
            moon_id=1
        )
        self.rs2 = Resource.objects.create(
            ore_id=2,
            quantity=0.1945375,
            moon_id=1
        )
        self.rs3 = Resource.objects.create(
            ore_id=3,
            quantity=0.1866783,
            moon_id=1
        )
        self.rs4 = Resource.objects.create(
            ore_id=4,
            quantity=0.15992839,
            moon_id=1
        )
        self.rs5 = Resource.objects.create(
            ore_id=5,
            quantity=0.168848493,
            moon_id=1
        )
        self.rs6 = Resource.objects.create(
            ore_id=6,
            quantity=0.00003,
            moon_id=1
        )

        self.order = [self.rs1, self.rs2, self.rs3, self.rs5, self.rs4, self.rs6]

        # Refinery Data
        EveCategory.objects.create(id=10, published=True)
        EveGroup.objects.create(id=10, published=True, eve_category_id=10)
        EveType.objects.create(id=10, published=True, eve_group_id=10, name="Type")
        EveCorporationInfo.objects.create(corporation_name="Test",
                                          corporation_ticker="ABC",
                                          corporation_id=123,
                                          member_count=1
                                          )

        self.r1 = Refinery.objects.create(
            structure_id=1,
            evetype_id=1,
            name="Test Structure",
            corp_id=1
        )

        self.r2 = Refinery.objects.create(
            structure_id=2,
            evetype_id=1,
            name="Test Structure",
            corp_id=1
        )

        # Extraction Data
        # True
        self.ex1 = Extraction.objects.create(
            start_time=datetime(1970, 1, 1),
            arrival_time=datetime(1970, 1, 2),
            decay_time=datetime(1970, 1, 3),
            moon_id=1,
            refinery_id=1,
            corp_id=1,
            active=False
        )

        # False
        self.ex2 = Extraction.objects.create(
            start_time=datetime(1970, 1, 2).replace(tzinfo=pytz.utc),
            arrival_time=datetime(1970, 1, 3).replace(tzinfo=pytz.utc),
            decay_time=datetime(1970, 1, 4).replace(tzinfo=pytz.utc),
            moon_id=1,
            refinery_id=1,
            corp_id=1,
            active=True
        )

        # True
        self.ex3 = Extraction.objects.create(
            start_time=datetime.utcnow().replace(tzinfo=pytz.utc),
            arrival_time=datetime.utcnow().replace(tzinfo=pytz.utc) + timedelta(days=1),
            decay_time=datetime.utcnow().replace(tzinfo=pytz.utc) + timedelta(days=2),
            moon_id=1,
            refinery_id=1,
            corp_id=1,
            depleted=True
        )

        # False
        self.ex4 = Extraction.objects.create(
            start_time=datetime.utcnow().replace(tzinfo=pytz.utc),
            arrival_time=datetime.utcnow().replace(tzinfo=pytz.utc) + timedelta(days=2),
            decay_time=datetime.utcnow().replace(tzinfo=pytz.utc) + timedelta(days=3),
            moon_id=1,
            refinery_id=1,
            corp_id=1,
            active=False
        )

        self.ex1 = Extraction.objects.create(
            start_time=datetime(1970, 1, 1),
            arrival_time=datetime(1970, 1, 2),
            decay_time=datetime(1970, 1, 3),
            moon_id=3,
            refinery_id=2,
            corp_id=1,
            active=False
        )

    def test_get_refinery_name(self):
        """
        Test the get_refinery_name filter.
        :return:
        """
        name = filters.get_refinery_name(self.moon)
        name2 = filters.get_refinery_name(self.moon2)
        self.assertEqual("Test Structure", name)
        self.assertEqual("", name2)

    def test_get_refinery_owner_name(self):
        """
        Test the get_refinery_owner_name filter.
        :return:
        """
        name = filters.get_refinery_owner_name(self.moon)
        name2 = filters.get_refinery_owner_name(self.moon2)
        self.assertEqual("Test", name)
        self.assertEqual("", name2)

    def test_get_refinery_owner_id(self):
        """
        Test the get_refinery_owner_id filter
        :return:
        """
        corp_id = filters.get_refinery_owner_id(self.moon)
        corp_id2 = filters.get_refinery_owner_id(self.moon2)
        self.assertEqual(123, corp_id)
        self.assertEqual('', corp_id2)

    def test_get_next_extraction(self):
        """
        Test the get_next_extraction filter
        :return:
        """
        ext = filters.get_next_extraction(self.moon)
        ext2 = filters.get_next_extraction(self.moon2)
        ext3 = filters.get_next_extraction(self.moon3)
        self.assertEqual(ext, datetime.strftime(self.ex4.arrival_time, '%Y-%m-%d %H:%M'))
        self.assertEqual(ext2, '')
        self.assertEqual(ext3, '')

    def test_check_visibility(self):
        """
        Test the check_visibility filter.
        :return:
        """
        self.assertTrue(filters.check_visibility(self.ex1))
        self.assertFalse(filters.check_visibility(self.ex2))
        self.assertTrue(filters.check_visibility(self.ex3))
        self.assertFalse(filters.check_visibility(self.ex4))

    def test_card_labels(self):
        """
        Test the card_labels filter.
        :return:
        """
        labels = filters.card_labels(self.moon.resources.all())
        self.assertEqual(labels, [64, 32, 16, 8, 4, 0])

    def test_chunk_time(self):
        """
        Test the chunk_time filter.
        :return:
        """
        time = filters.chunk_time(self.ex4)
        self.assertEqual(2, time)

    def test_percent(self):
        """
        Test the percent filter.
        :return:
        """
        pct = filters.percent(0.25)
        self.assertEqual(pct, 25)

    def test_order_quantity(self):
        """
        Test the order_quantity filter.
        :return:
        """
        res = filters.order_quantity(self.moon.resources.all())
        self.assertEqual(res, self.order)

    def test_get_item(self):
        """
        Test the get_item filter.
        :return:
        """
        dct = {'a': 32, 'b': 9}
        a = filters.get_item(dct, 'a')
        n = filters.get_item(dct, 'c')
        self.assertEqual(a, 32)
        self.assertEqual(n, None)




from unittest import mock
from datetime import datetime
from django.test import TestCase
from allianceauth.tests.auth_utils import AuthUtils
from esi.models import Token, Scope
from allianceauth.eveonline.models import EveCharacter
from eveuniverse.models import EveMoon, EvePlanet, EveSolarSystem, EveConstellation, EveType, EveGroup, EveCategory,\
    EveRegion


from ..tasks import filetime_to_dt, _get_tokens, _get_corp_tokens, process_scan
from ..models import TrackingCharacter, Resource
from .parser_strings import with_header


class TestTasks(TestCase):
    def setUp(self):
        self.expected_dt = datetime(1970, 1, 1)

        self.user1 = AuthUtils.create_user("Test User1")
        self.char1 = EveCharacter.objects.create(
            character_name="The First Char",
            character_id=1,
            corporation_name="The First Corp",
            corporation_id=123,
            corporation_ticker="ABC"
        )

        self.user1.profile.main_character = self.char1
        self.user1.profile.save()
        self.tracking1 = TrackingCharacter.objects.create(
            character=self.char1
        )
        self.token = Token.objects.create(
            access_token='access',
            refresh_token='refresh',
            user=self.user1,
            character_id=1,
            character_name='The First Char',
            token_type='Character',
            character_owner_hash='abcde',
        )
        scope, _ = Scope.objects.get_or_create(
            name="scope.v1"
        )
        self.token.scopes.add(scope)

        # EveUniverse Models
        EveCategory.objects.create(id=1, published=True)
        EveCategory.objects.create(id=2, published=True)
        EveGroup.objects.create(id=1884, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1921, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1, published=True, eve_category_id=1)
        EveType.objects.create(id=1, published=True, eve_group_id=1)
        EveType.objects.create(id=45493, published=True, eve_group_id=1884)
        EveType.objects.create(id=45499, published=True, eve_group_id=1921)
        EveType.objects.create(id=45490, published=True, eve_group_id=1884)
        EveRegion.objects.create(id=1)
        EveConstellation.objects.create(id=1, eve_region_id=1)
        EveSolarSystem.objects.create(id=1, security_status=1, eve_constellation_id=1)
        EvePlanet.objects.create(id=1, eve_solar_system_id=1, eve_type_id=1)
        EveMoon.objects.create(id=40217116, eve_planet_id=1)

    def test_filetime_conversion(self):
        """
        Tests the function that converts filetime (LDAP time) to datetime
        :return:
        """
        ft = 116444736000000000
        dt = filetime_to_dt(ft)
        self.assertEqual(dt, self.expected_dt)

    def test_get_tokens(self):
        """
        Tests the _get_tokens function.
        :return:
        """
        tokens = _get_tokens(["scope.v1", ])
        self.assertEqual(tokens, [self.token, ])

    def test_get_corp_tokens(self):
        """
        Tests the _get_corp_tokens function.
        :return:
        """
        tokens = _get_corp_tokens(123, ["scope.v1", ])
        self.assertEqual(tokens, [self.token, ])

    def test_process_scan(self):
        """
        Tests the process_scan task.
        :return:
        """
        process_scan(with_header, 1)
        self.assertTrue(EveMoon.objects.get(id=40217116))
        self.assertTrue(len(Resource.objects.filter(moon__id=40217116)) == 3)



# Generated by Django 3.1.2 on 2021-05-15 03:37

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0014_auto_20210105_1413'),
        ('moonstuff', '0004_delete_trackingcharacter'),
    ]

    operations = [
        migrations.CreateModel(
            name='TrackingCharacter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('latest_notification_id', models.BigIntegerField(default=0, null=True)),
                ('last_notification_check', models.DateTimeField(null=True)),
                ('character', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='eveonline.evecharacter')),
            ],
            options={
                'default_permissions': ('add',),
            },
        ),
    ]



from unittest import mock
from django.test import TestCase
from django.db.utils import IntegrityError
from eveuniverse.models import EveMoon, EvePlanet, EveSolarSystem, EveConstellation, EveType, EveGroup, EveCategory,\
    EveRegion
from allianceauth.eveonline.models import EveCorporationInfo, EveCharacter
from datetime import datetime

from ..models import Resource, Refinery, TrackingCharacter, Extraction, LedgerEntry


class TestResource(TestCase):
    def setUp(self):
        EveCategory.objects.create(id=1, published=True)
        EveCategory.objects.create(id=2, published=True)
        EveGroup.objects.create(id=1884, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1920, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1921, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1922, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1923, published=True, eve_category_id=2)
        EveGroup.objects.create(id=1, published=True, eve_category_id=1)
        EveType.objects.create(id=1, published=True, eve_group_id=1884)
        EveType.objects.create(id=2, published=True, eve_group_id=1920)
        EveType.objects.create(id=3, published=True, eve_group_id=1921)
        EveType.objects.create(id=4, published=True, eve_group_id=1922)
        EveType.objects.create(id=5, published=True, eve_group_id=1923)
        EveType.objects.create(id=6, published=True, eve_group_id=1)
        EveRegion.objects.create(id=1)
        EveConstellation.objects.create(id=1, eve_region_id=1)
        EveSolarSystem.objects.create(id=1, security_status=1, eve_constellation_id=1)
        EvePlanet.objects.create(id=1, eve_solar_system_id=1, eve_type_id=6)
        EveMoon.objects.create(id=1, eve_planet_id=1)

        Resource.objects.create(
            ore_id=1,
            quantity=0.234445456,
            moon_id=1
        )
        Resource.objects.create(
            ore_id=2,
            quantity=0.1945375,
            moon_id=1
        )
        Resource.objects.create(
            ore_id=3,
            quantity=0.1866783,
            moon_id=1
        )
        Resource.objects.create(
            ore_id=4,
            quantity=0.15992839,
            moon_id=1
        )
        Resource.objects.create(
            ore_id=5,
            quantity=0.168848493,
            moon_id=1
        )
        Resource.objects.create(
            ore_id=6,
            quantity=0.00003,
            moon_id=1
        )

    def test_rarity(self):
        res = Resource.objects.filter(moon_id=1)
        self.assertTrue(len(res) == 6)
        self.assertEqual(res[0].rarity, 4)
        self.assertEqual(res[1].rarity, 8)
        self.assertEqual(res[2].rarity, 16)
        self.assertEqual(res[3].rarity, 32)
        self.assertEqual(res[4].rarity, 64)
        self.assertEqual(res[5].rarity, 0)


class TestRefinery(TestCase):
    def setUp(self):
        EveCategory.objects.create(id=1, published=True)
        EveGroup.objects.create(id=1, published=True, eve_category_id=1)
        EveType.objects.create(id=1, published=True, eve_group_id=1, name="Type")
        EveCorporationInfo.objects.create(corporation_name="Test",
                                          corporation_ticker="ABC",
                                          corporation_id=123,
                                          member_count=1
                                          )

        self.r1 = Refinery.objects.create(
            structure_id=1,
            evetype_id=1,
            name="Test Structure",
            corp_id=1
        )
        self.r2 = Refinery.objects.create(
            structure_id=2,
            evetype_id=1,
            corp_id=1
        )

    def test_str_with_name(self):
        """
        Test __str__ method
        :return:
        """
        self.assertEqual(self.r1.__str__(), "Test Structure")

    def test_str_with_missing_name(self):
        """
        Test the __str__ method on an object missing a name.
        :return:
        """
        self.assertEqual(self.r2.__str__(), "Unknown Structure ID2 (Type)")


class TestTrackingCharacter(TestCase):
    def setUp(self):
        self.character = EveCharacter.objects.create(
            character_id=1,
            character_name="Test Character",
            corporation_id=123,
            corporation_name="Test Corp",
            corporation_ticker="ABC"
        )

        self.tracker = TrackingCharacter.objects.create(
            character=self.character
        )

    def test_str(self):
        """
        Test the __str__ method
        :return:
        """
        self.assertEqual(self.tracker.__str__(), "Test Character")

    def test_only_one_trackingcharacter_per_character(self):
        """
        Test the OtO relationship between EveCharacter and TrackingCharacter.
        :return:
        """
        self.assertRaises(IntegrityError, TrackingCharacter.objects.create, character=self.character)


class TestExtraction(TestCase):
    def setUp(self):
        EveCategory.objects.create(id=1, published=True)
        EveCategory.objects.create(id=2, published=True)
        EveGroup.objects.create(id=1, published=True, eve_category_id=1)
        EveGroup.objects.create(id=2, published=True, eve_category_id=1)
        EveType.objects.create(id=1, published=True, eve_group_id=1)
        EveType.objects.create(id=2, published=True, eve_group_id=1, name="Type")
        EveRegion.objects.create(id=1)
        EveConstellation.objects.create(id=1, eve_region_id=1)
        EveSolarSystem.objects.create(id=1, security_status=1, eve_constellation_id=1)
        EvePlanet.objects.create(id=1, eve_solar_system_id=1, eve_type_id=1)
        EveMoon.objects.create(id=1, eve_planet_id=1)

        EveCorporationInfo.objects.create(corporation_name="Test",
                                          corporation_ticker="ABC",
                                          corporation_id=123,
                                          member_count=1
                                          )

        self.r1 = Refinery.objects.create(
            structure_id=1,
            evetype_id=2,
            name="Test Structure",
            corp_id=1
        )

        self.extraction_values = {
            'start_time': datetime(1970, 1, 1),
            'arrival_time': datetime(1970, 1, 2),
            'decay_time': datetime(1970, 1, 3),
            'moon_id': 1,
            'refinery_id': 1,
            'corp_id': 1
        }

        self.extraction = Extraction.objects.create(
            **self.extraction_values
        )

    def test_despawn_property(self):
        """
        Test the despawn property of the Extraction class.
        :return:
        """
        self.assertEqual(self.extraction.despawn, datetime(1970, 1, 5))

    def test_unique_start_and_moon(self):
        """
        Test the unique_together condition
        :return:
        """
        self.assertRaises(IntegrityError, Extraction.objects.create, **self.extraction_values)


class TestLedgerEntry(TestCase):
    def setUp(self):
        EveCategory.objects.create(id=1, published=True)
        EveCategory.objects.create(id=2, published=True)
        EveGroup.objects.create(id=1, published=True, eve_category_id=1)
        EveGroup.objects.create(id=2, published=True, eve_category_id=2)
        EveType.objects.create(id=1, published=True, eve_group_id=1, name="Type")
        EveType.objects.create(id=2, published=True, eve_group_id=2, name="Non-Refinery Type")
        EveCorporationInfo.objects.create(corporation_name="Test",
                                          corporation_ticker="ABC",
                                          corporation_id=123,
                                          member_count=1
                                          )

        self.r1 = Refinery.objects.create(
            structure_id=1,
            evetype_id=1,
            name="Test Structure",
            corp_id=1
        )

        self.ledger_entry_values = {
            'observer': self.r1,
            'character_id': 1,
            'last_updated': datetime(1970, 1, 1),
            'quantity': 123456789,
            'recorded_corporation_id': 1,
            'evetype_id': 2
        }

        LedgerEntry.objects.create(**self.ledger_entry_values)

    def test_unique_together(self):
        """
        Test the unique_together condition
        :return:
        """
        self.assertRaises(IntegrityError, LedgerEntry.objects.create, **self.ledger_entry_values)



from django.contrib import admin

from .models import Extraction, Refinery, TrackingCharacter, Resource


# Register your models here.
class TrackingCharacterAdmin(admin.ModelAdmin):
    list_display = ("character",
                    "get_corp",
                    "latest_notification_id",
                    "last_notification_check")

    def get_corp(self, obj):
        return obj.character.corporation_name

    get_corp.short_description = "Corporation"
    get_corp.admin_order_field = "character__corporation_name"

    list_filter = ("character__corporation_name",)
    search_fields = ("character__corporation_name", "character__character_name")


class ExtractionAdmin(admin.ModelAdmin):
    list_select_related = True
    list_display = ("moon",
                    "refinery",
                    "corp",
                    "cancelled",
                    "start_time",
                    "arrival_time",
                    "total_volume",
                    "decay_time",
                    "despawn",
                    "active",
                    "jackpot",
                    "depleted")

    list_filter = ("moon", "corp", "refinery", "jackpot", "cancelled", "active")
    search_fields = ("corp__corporation_name", "moon__name", "refinery__name")


class EveTypeFilter(admin.SimpleListFilter):
    title = "Structure Type"
    parameter_name = "type"

    def lookups(self, request, model_admin):
        types = set([r.evetype for r in model_admin.model.objects.all()])
        return [(t.id, t.name) for t in types]

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(evetype__id__exact=self.value())


class RefineryAdmin(admin.ModelAdmin):

    list_select_related = True
    list_display = ("name",
                    "evetype",
                    "corp",
                    "structure_id",
                    "observer")
    list_filter = (EveTypeFilter, "corp", "observer")
    search_fields = ("corp__corporation_name", "evetype__name", "name")


class EveMoonFilter(admin.SimpleListFilter):
    title = "Moon"
    parameter_name = "moon"

    def lookups(self, request, model_admin):
        moons = set([r.moon for r in model_admin.model.objects.all()])
        return [(m.id, m.name) for m in moons]

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(moon__id__exact=self.value())


class OreFilter(admin.SimpleListFilter):
    title = "Ore"
    parameter_name = "ore"

    def lookups(self, request, model_admin):
        ore = set([r.ore for r in model_admin.model.objects.all()])
        return [(o.id, o.name) for o in ore]

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(ore__id__exact=self.value())


class ResourceAdmin(admin.ModelAdmin):
    list_select_related = True
    list_display = ("moon",
                    "ore",
                    "quantity")
    list_filter = (EveMoonFilter, OreFilter)
    search_fields = ("moon", "ore")


admin.site.register(Resource, ResourceAdmin)
admin.site.register(Refinery, RefineryAdmin)
admin.site.register(Extraction, ExtractionAdmin)
admin.site.register(TrackingCharacter, TrackingCharacterAdmin)



# Generated by Django 3.1.2 on 2021-06-17 02:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('moonstuff', '0005_trackingcharacter'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='refinery',
            options={'default_permissions': (), 'verbose_name_plural': 'Refineries'},
        ),
    ]



from django.template.defaulttags import register
from django.utils import timezone
from datetime import datetime, timedelta
import pytz


@register.filter()
def get_refinery_name(moon):
    exts = moon.extractions.all()
    if len(exts) == 0:
        return ''
    refinery = list(exts)[-1].refinery
    return refinery.name


@register.filter()
def get_refinery_owner_name(moon):
    exts = moon.extractions.all()
    if len(exts) == 0:
        return ''
    refinery = list(exts)[-1].refinery
    return refinery.corp.corporation_name


@register.filter()
def get_refinery_owner_id(moon):
    exts = moon.extractions.all()
    if len(exts) == 0:
        return ''
    refinery = list(exts)[-1].refinery
    return refinery.corp.corporation_id


@register.filter()
def get_next_extraction(moon):
    """
    Returns the next extraction.
    This assumes that the last extraction listed for a moon is the next one, as you cant schedule two at once.
    :param moon:
    :return:
    """
    exts = list(moon.extractions.all())
    if len(exts) == 0:
        return ''
    ext_arrival = exts[-1].arrival_time
    now = timezone.now()
    if ext_arrival > now:
        return datetime.strftime(exts[-1].arrival_time, '%Y-%m-%d %H:%M')
    return ''


@register.filter()
def check_visibility(extraction):
    """
    Returns true if extraction should *not* be visible.
    :param extraction:
    :return:
    """
    return (not extraction.active and
            datetime.utcnow().replace(tzinfo=pytz.utc) > extraction.despawn.replace(tzinfo=pytz.utc)
            ) or extraction.depleted


@register.filter()
def card_labels(resources):
    rare_values = set([r.rarity for r in resources])
    return sorted(list(rare_values), reverse=True)


@register.filter()
def chunk_time(extraction):
    """
    Returns the number of days that the extraction will take.
    :param extraction: Extraction model object.
    :return:
    """
    start = extraction.start_time
    end = extraction.arrival_time

    delta = end - start
    return delta.days


@register.filter()
def percent(quantity: float):
    """
    Converts decimal to percent.
    :param quantity: float
    :return:
    """
    return round(quantity * 100, 1)


@register.filter()
def order_quantity(resources):
    """
    Returns a list of resources ordered by quantity.
    :param resources: QS containing the resources.
    :return:
    """
    return sorted(list(resources), key=lambda r: r.quantity, reverse=True)


@register.filter()
def get_item(dictionary, key):
    return dictionary.get(key, None)



from django.urls import re_path

from . import views

app_name = "moonstuff"

urlpatterns = [
    re_path(r'^$', views.dashboard, name='dashboard'),
    re_path(r'^scan/$', views.add_scan, name='add_scan'),
    re_path(r'^track/$', views.add_character, name='add_character'),
    re_path(r'^info/(?P<moon_id>[0-9]+)/$', views.moon_info, name='view_moon')
]



from esi.clients import EsiClientProvider

from . import __version__

ESI_CHARACTER_SCOPES = (
    'esi-industry.read_corporation_mining.v1',
    'esi-universe.read_structures.v1',
    'esi-characters.read_notifications.v1',
    )

esi = EsiClientProvider(app_info_text="moonstuff v" + __version__)



# Generated by Django 3.1.2 on 2021-04-18 22:09

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('eveonline', '0014_auto_20210105_1413'),
        ('eveuniverse', '0004_effect_longer_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Moonstuff',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': (('access_moonstuff', 'Allows access to the moonstuff module'), ('access_moon_list', 'Allows access to list of all moons.')),
                'managed': False,
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='MaterialCheckSum',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('checksum', models.CharField(default=None, max_length=255, null=True)),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='TrackingCharacter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('latest_notification_id', models.BigIntegerField(default=0, null=True)),
                ('last_notification_check', models.DateTimeField(null=True)),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveonline.evecharacter')),
            ],
            options={
                'default_permissions': ('add',),
            },
        ),
        migrations.CreateModel(
            name='Resource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.DecimalField(decimal_places=10, max_digits=11)),
                ('moon', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='resources', to='eveuniverse.evemoon')),
                ('ore', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype')),
            ],
            options={
                'default_permissions': ('add',),
            },
        ),
        migrations.CreateModel(
            name='Refinery',
            fields=[
                ('structure_id', models.BigIntegerField(primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255, null=True)),
                ('observer', models.BooleanField(default=True)),
                ('corp', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='refineries', to='eveonline.evecorporationinfo')),
                ('evetype', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Material',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.IntegerField()),
                ('evetype', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='materials', to='eveuniverse.evetype')),
                ('material_evetype', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='from_types', to='eveuniverse.evetype')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='LedgerEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('character_id', models.IntegerField()),
                ('last_updated', models.DateField()),
                ('quantity', models.BigIntegerField()),
                ('recorded_corporation_id', models.IntegerField()),
                ('evetype', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='ledger_entries', to='eveuniverse.evetype')),
                ('observer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='moonstuff.refinery')),
            ],
            options={
                'default_permissions': (),
                'unique_together': {('observer', 'last_updated', 'character_id', 'evetype')},
            },
        ),
        migrations.CreateModel(
            name='Extraction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_time', models.DateTimeField()),
                ('arrival_time', models.DateTimeField()),
                ('decay_time', models.DateTimeField()),
                ('cancelled', models.BooleanField(default=False)),
                ('jackpot', models.BooleanField(default=False)),
                ('active', models.BooleanField(default=False)),
                ('depleted', models.BooleanField(default=False)),
                ('total_volume', models.BigIntegerField(null=True)),
                ('corp', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extractions', to='eveonline.evecorporationinfo')),
                ('moon', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extractions', to='eveuniverse.evemoon')),
                ('refinery', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extractions', to='moonstuff.refinery')),
            ],
            options={
                'default_permissions': (),
                'unique_together': {('start_time', 'moon')},
            },
        ),
    ]



# Generated by Django 3.1.2 on 2021-05-15 03:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('moonstuff', '0003_OtO_TrackingCharacter_Character'),
    ]

    operations = [
        migrations.DeleteModel(
            name='TrackingCharacter',
        ),
    ]



from django.db import models
from datetime import timedelta


from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from eveuniverse.models import EveMoon, EveType


# Create your models here.
class Moonstuff(models.Model):

    class Meta:
        managed = False
        default_permissions = (())
        permissions = (
            ('access_moonstuff', 'Allows access to the moonstuff module'),
            ('access_moon_list', 'Allows access to list of all moons.'),
        )


class Resource(models.Model):
    ore = models.ForeignKey(EveType, on_delete=models.CASCADE)
    quantity = models.DecimalField(max_digits=11, decimal_places=10)
    moon = models.ForeignKey(EveMoon, on_delete=models.CASCADE, related_name='resources')

    @property
    def rarity(self):
        """
        Determine the rarity of the resource by its group_id. (Default is 0)
        :return:
        """
        rs = {1884: 4, 1920: 8, 1921: 16, 1922: 32, 1923: 64}
        return rs.get(self.ore.eve_group_id, 0)

    def __str__(self):
        return f"{self.ore.name} x {self.quantity} ({self.moon.name})"

    class Meta:
        default_permissions = ('add',)


class Refinery(models.Model):
    structure_id = models.BigIntegerField(primary_key=True)
    evetype = models.ForeignKey(EveType, on_delete=models.CASCADE)
    name = models.CharField(null=True, max_length=255)  # Might not actually need this.
    corp = models.ForeignKey(EveCorporationInfo, on_delete=models.CASCADE, related_name='refineries')
    observer = models.BooleanField(default=True)

    def __str__(self):
        if self.name is None:
            return f'Unknown Structure ID{self.structure_id} ({self.evetype.name})'
        return f'{self.name}'

    class Meta:
        default_permissions = (())
        verbose_name_plural = "Refineries"


class TrackingCharacter(models.Model):
    character = models.OneToOneField(EveCharacter, on_delete=models.CASCADE)
    latest_notification_id = models.BigIntegerField(null=True, default=0)
    last_notification_check = models.DateTimeField(null=True)

    def __str__(self):
        return f'{self.character.character_name}'

    class Meta:
        default_permissions = ('add',)


class Extraction(models.Model):
    start_time = models.DateTimeField()
    arrival_time = models.DateTimeField()
    decay_time = models.DateTimeField()
    moon = models.ForeignKey(EveMoon, on_delete=models.CASCADE, related_name='extractions')
    refinery = models.ForeignKey(Refinery, on_delete=models.CASCADE, related_name='extractions')
    corp = models.ForeignKey(EveCorporationInfo, on_delete=models.CASCADE, related_name='extractions')
    cancelled = models.BooleanField(null=False, default=False)
    jackpot = models.BooleanField(default=False)
    active = models.BooleanField(default=False)
    depleted = models.BooleanField(default=False)
    total_volume = models.BigIntegerField(null=True)

    @property
    def despawn(self):
        """
        Returns the latest despawn time. Actual despawn can be up to -3h from returned time depending on when the
            moon drill laser was fired. Actual despawn should never occur after returned time however.
        :return:
        """
        return self.decay_time + timedelta(hours=48)

    class Meta:
        default_permissions = (())
        unique_together = (('start_time', 'moon'),)


class LedgerEntry(models.Model):
    observer = models.ForeignKey(Refinery, on_delete=models.CASCADE, related_name='entries')
    character_id = models.IntegerField()  # There is no guarantee that we will have a character object for this.
    last_updated = models.DateField()
    quantity = models.BigIntegerField()
    recorded_corporation_id = models.IntegerField()
    evetype = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name='ledger_entries')

    class Meta:
        default_permissions = (())
        # For each observer we only want one record per day per character per ore type.
        unique_together = (('observer', 'last_updated', 'character_id', 'evetype'),)



# Generated by Django 3.1.2 on 2021-04-19 06:44

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('moonstuff', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='material',
            name='evetype',
        ),
        migrations.RemoveField(
            model_name='material',
            name='material_evetype',
        ),
        migrations.DeleteModel(
            name='MaterialCheckSum',
        ),
        migrations.DeleteModel(
            name='Material',
        ),
    ]



import requests
import yaml
import datetime
import pytz

from allianceauth.services.hooks import get_extension_logger
from allianceauth.notifications import notify
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from celery import shared_task
from eveuniverse.tasks import update_or_create_eve_object
from eveuniverse.models import EveUniverseEntityModel, EveMarketPrice
from django.db.models import Q, Count, Sum, F, BigIntegerField
from django.db.utils import IntegrityError
from django.contrib.auth.models import User
from django.utils.translation import gettext as gt
from esi.models import Token

from .providers import esi, ESI_CHARACTER_SCOPES
from .models import \
    EveType, Resource, EveMoon, TrackingCharacter, Refinery, Extraction, LedgerEntry
from .parser import ScanParser

logger = get_extension_logger(__name__)


def _get_tokens(scopes):
    """
    Gets all tokens with matching scopes.
    :param scopes:
    :return:
    """
    try:
        tokens = list()
        characters = TrackingCharacter.objects.all()
        for character in characters:
            tokens.append(Token.get_token(character.character.character_id, scopes))
        return tokens
    except Exception as e:
        print(e)
        return False


def _get_corp_tokens(corp_id, scopes):
    """
    Gets all tokens with matching corp and scopes.
    :param corp_id: integer
    :param scopes: list(String)
    :return:
    """
    try:
        tokens = list()
        characters = TrackingCharacter.objects.filter(character__corporation_id=corp_id)
        for character in characters:
            tokens.append(Token.get_token(character.character.character_id, scopes))
        return tokens
    except Exception as e:
        print(e)
        return False


def filetime_to_dt(ft):
    us = (ft - 116444736000000000) // 10
    return datetime.datetime(1970, 1, 1) + datetime.timedelta(microseconds=us)


@shared_task()
def load_types_and_mats(category_ids=None, group_ids=None, type_ids=None, force_loading_dogma=False):
    logger.debug(f'Calling eveuniverse load functions for the following args:'
                 f' cats: {category_ids}'
                 f' groups: {group_ids}'
                 f' types: {type_ids}'
                 f' dogma? {force_loading_dogma}')

    # Synchronously load SDE data to ensure it exists before we spawn the material loading task.
    # This is basically the logic for the _load_eve_xxx functions from eveuniverse thrown in here to force synchronicity
    if category_ids:
        for category_id in category_ids:
            enabled_sections = (
                [EveUniverseEntityModel.LOAD_DOGMAS] if force_loading_dogma else None
            )
            update_or_create_eve_object(
                model_name="EveCategory",
                id=category_id,
                include_children=True,
                wait_for_children=False,
                enabled_sections=enabled_sections,
            )

    if group_ids:
        for group_id in group_ids:
            enabled_sections = (
                [EveUniverseEntityModel.LOAD_DOGMAS] if force_loading_dogma else None
            )
            update_or_create_eve_object(
                model_name="EveGroup",
                id=group_id,
                include_children=True,
                wait_for_children=False,
                enabled_sections=enabled_sections,
            )

    if type_ids:
        for type_id in type_ids:
            enabled_sections = (
                [EveUniverseEntityModel.LOAD_DOGMAS] if force_loading_dogma else None
            )
            update_or_create_eve_object(
                model_name="EveType",
                id=type_id,
                include_children=False,
                wait_for_children=False,
                enabled_sections=enabled_sections,
            )

    logger.debug('Done loading eve types! Scheduling price loading.')
    # Any time types are loaded we should ensure we have material and price data for all types
    load_prices.delay()


@shared_task()
def process_scan(scan_data: str, user_id: int):
    """
    Runs the provided scan data through the parser, and creates the required resource objects.
    :param scan_data: The raw scan data from the view.
    :param user_id: The user that initiated the scan.
    :return:
    """
    logger.debug('Processing moon scan.')
    try:
        data = ScanParser(scan_data).parse()

        resources = list()

        for moon in data:
            moon, created = EveMoon.objects.get_or_create_esi(id=int(moon))
            if not created:
                Resource.objects.filter(moon=moon).delete()

        for res_l in data.values():
            for res in res_l:
                resources.append(Resource(**res))

        Resource.objects.bulk_create(resources)
        logger.debug("Successfully processed moon scan!")
    except Exception as e:
        logger.error(f'Failed processing moon scan! (Data sent to user_id {user_id} via notification)')
        notify(
            User.objects.get(id=user_id),
            gt('Failed to Process Moon Scan'),
            message=gt('There was an error processing the following moon scan:\n'
                       '%(scan)s'
                       '\n\n'
                       'Error Encountered: %(error)s\n') % {'scan': scan_data, 'error': e},
            level='danger'
        )


@shared_task()
def load_prices():
    """
    Updates EveMarketPrice records.
    :return:
    """
    EveMarketPrice.objects.update_from_esi()


@shared_task()
def import_extraction_data():
    """
    Imports extraction data, and schedules notification checks.
    :return:
    """
    client = esi.client
    tokens = _get_tokens(ESI_CHARACTER_SCOPES)
    for token in tokens:
        try:
            # Get character and corp objects to go with token
            char = EveCharacter.objects.get(character_id=token.character_id)
            try:
                corp = EveCorporationInfo.objects.get(corporation_id=char.corporation_id)
            except EveCorporationInfo.DoesNotExist:
                corp = EveCorporationInfo.objects.create_corporation(corp_id=char.corporation_id)

            # Get Extraction events.
            events = client.Industry.get_corporation_corporation_id_mining_extractions(
                corporation_id=corp.corporation_id,
                token=token.valid_access_token()
            ).results()

            for event in events:
                # Get Structure Info
                moon, _ = EveMoon.objects.get_or_create_esi(id=event['moon_id'])
                try:
                    refinery = Refinery.objects.get(structure_id=event['structure_id'])
                except Refinery.DoesNotExist:
                    ref = client.Universe.get_universe_structures_structure_id(
                        structure_id=event['structure_id'],
                        token=token.valid_access_token()
                    ).results()
                    refinery = Refinery(
                        structure_id=event['structure_id'],
                        name=ref['name'],
                        corp=corp,
                        evetype_id=ref['type_id']
                    )
                    refinery.save()

                start_time = event['extraction_start_time']
                arrival_time = event['chunk_arrival_time']
                decay_time = event['natural_decay_time']

                # Calculate the total volume for the extraction. (40k m3 per hour)
                total_volume = ((arrival_time - start_time) / datetime.timedelta(seconds=3600)) * 40000

                try:
                    # Create the extraction event.
                    extraction = Extraction.objects.get_or_create(
                        start_time=start_time,
                        arrival_time=arrival_time,
                        decay_time=decay_time,
                        refinery=refinery,
                        moon=moon,
                        corp=corp,
                        total_volume=total_volume,
                    )
                except IntegrityError:
                    continue
                except Exception as e:
                    logger.error(f'Error encountered when saving extraction! Corp ID: {corp.corporation_id}'
                                 f' Refinery ID: {refinery.structure_id} Event Start: {start_time}')
                    logger.error(e)
            logger.info(f'Imported extraction data from {token.character_id}')
        except Exception as e:
            logger.error(f'Error importing extraction data from {token.character_id}')
            logger.error(e)

        check_notifications.delay(token.character_id)


@shared_task()
def check_notifications(character_id: int):
    """
    Checks and processes notifications related to moon mining.
        Note: This task does not add extraction events!
    :param character_id: The character_id to use to get a token.
    :return:
    """
    logger.debug(f'Checking notifications for {character_id}')
    # Define token and client, ensuring the token is valid.
    client = esi.client
    token = Token.get_token(character_id, ESI_CHARACTER_SCOPES)
    char = EveCharacter.objects.get(character_id=token.character_id)
    char = TrackingCharacter.objects.get(character=char)
    last_noti = char.latest_notification_id

    # Get notifications
    notifications = client.Character.get_characters_character_id_notifications(
        character_id=char.character.character_id,
        token=token.valid_access_token()
    ).results()
    # Set the last notification id for the character
    notifications.reverse()  # We want the newest data last... so reverse the list
    char.latest_notification_id = notifications[-1]['notification_id']
    char.save()

    # Filter out notifications that we dont care about
    notifications = list(
        filter(
            lambda n: 'Moonmining' in n['type'] and int(n['notification_id']) > last_noti,
            notifications
        )
    )

    # Start processing notifications
    for noti in notifications:
        if 'Cancelled' not in noti['type']:
            # First parse the text from yaml format.
            data = yaml.safe_load(noti['text'])

            # Check that the moon has resources associated with it.
            # (If a scan was never added, it might not)
            moon, created = EveMoon.objects.get_or_create_esi(id=data['moonID'])

            # Select the relevant extraction if applicable
            if 'Finished' in noti['type']:
                # We have decay time
                extractions = Extraction.objects.filter(
                    decay_time=filetime_to_dt(data['autoTime']).replace(tzinfo=pytz.utc),
                    cancelled=False,
                    moon=moon
                )
            elif 'Fracture' in noti['type']:
                # We only have notification time
                # If the moon auto fractured then the decay time was at or before the notification, and the
                # arrival time was roughly 3 hours before that. (Allowing for a 5 minute window of error on notification
                # time)
                extractions = Extraction.objects.filter(
                    arrival_time__gte=noti['timestamp'] - datetime.timedelta(hours=3, minutes=5),
                    decay_time__lte=noti['timestamp'],
                    cancelled=False,
                    moon=moon
                )
            elif 'Fired' in noti['type']:
                # We only have notification time
                # If the laser was manually fired than it must be after the arrival time, but
                # before the decay time.
                extractions = Extraction.objects.filter(
                    arrival_time__lte=noti['timestamp'],
                    decay_time__gte=noti['timestamp'],
                    cancelled=False,
                    moon=moon
                )
            else:
                # We have arrival time
                logger.debug(f"ELSE {noti['type']} ID {noti['notification_id']}")
                extractions = Extraction.objects.filter(
                    arrival_time=filetime_to_dt(data['readyTime']).replace(tzinfo=pytz.utc),
                    cancelled=False,
                    moon=moon
                )

            if len(extractions) == 0 or len(extractions) > 1:
                extraction = None
            else:
                extraction = extractions[0]

            if created:
                # If the moon was created, then we don't know about the structure yet, so lets create it.
                owner = client.Universe.get_universe_structures_structure_id(
                    structure_id=data['structureID'],
                    token=token.valid_access_token()
                ).results()['owner_id']
                ref = Refinery(
                    structure_id=data['structureID'],
                    moon_id=data['moonID'],
                    evetype_id=data['structureTypeID'],
                    name=data['structureName'],
                    corp_id=owner
                )
                ref.save()
            res = moon.resources.all().values_list('ore_id', flat=True)
            missing_res = list()

            # Make a list of resources missing from the moon.
            # This is used in case the data is either incorrect or incomplete.
            for ore in data['oreVolumeByType']:
                if ore not in res:
                    missing_res.append(ore)

            # Set the active flag if the notification is either MoonminingAutomaticFracture or MoonminingLaserFired
            if ('AutomaticFracture' in noti['type'] or 'LaserFired' in noti['type']) and extraction is not None:
                extraction.active = True

            # Calculate the total volume of ore
            total_ore = 0
            for k, v in data['oreVolumeByType'].items():
                total_ore += v
            # Update the total volume of ore for the extraction
            if extraction is not None:
                extraction.total_volume = total_ore
                extraction.save()

            # If there is one or more missing resources, OR if there is a resource in the database
            # that shouldn't be there. We will assume that these notifications are always authoritative.
            if len(missing_res) > len(res) or len(missing_res) == len(data['oreVolumeByType']):
                # Calculate ore percentages, and add resource objects for the moon.
                # Create resource objects
                new_res = list()
                for k, v in data['oreVolumeByType'].items():
                    pct = v / total_ore
                    new_res.append(
                        Resource(moon=moon, ore_id=k, quantity=pct)
                    )

                # Delete old moon resources and create new ones.
                moon.resources.all().delete()
                Resource.objects.bulk_create(new_res)

        elif 'Cancelled' in noti['type']:
            # Determine which extraction event was cancelled and mark it as such.
            # First Parse the timestamp
            noti_time = datetime.datetime.strptime(noti['timestamp'], '%Y-%m-%dT%H:%M:%SZ')
            noti_time = pytz.timezone("UTC").localize(noti_time, is_dst=None)

            # Parse the notification data from yaml
            data = yaml.load(noti['text'])

            # Get the refinery and list of extractions
            try:
                refinery = Refinery.objects.get(structure_id=data['structureID'])
            except Refinery.DoesNotExist:
                logger.info(f'Got extraction cancellation notification for refinery not in database. '
                            f'NID {noti["notification_id"]}')
                continue

            exts = refinery.extractions.filter(start_time__lt=noti_time, arrival_time__gt=noti_time)\
                .order_by('-start_time')

            # Cancel the extraction(s).
            if len(exts) == 1:
                exts[0].cancelled = True
                exts[0].save()
            elif len(exts) > 1:
                for ext in exts:
                    if ext.cancelled is not True:
                        ext.cancelled = True
                        ext.save()
            else:
                logger.info(f'Got extraction cancellation notification for event not in database. '
                            f'NID {noti["notification_id"]}')


@shared_task()
def update_refineries():
    """
    Updates the names and observer status of all refineries.
    :return:
    """
    update_names.delay()
    update_observers.delay()


@shared_task()
def update_names():
    """
    Updates the names of refineries.
    :return:
    """

    client = esi.client

    corps = Refinery.objects.all().values_list('corp__corporation_id', flat=True)
    # Build a dict of tokens to try for each corp.
    tokens = dict()
    for corp in corps:
        ts = _get_corp_tokens(corp, ESI_CHARACTER_SCOPES)
        if ts:
            tokens[corp] = ts

    for corp in tokens:
        refs = Refinery.objects.filter(corp__corporation_id=corp)

        # If we have no tokens for this corp, we cant update the names... append '[STALE]' to structure names.
        if len(tokens[corp]) == 0:
            logger.info(f"No valid moon tracking tokens for CorpID: {corp}! Marking structures as STALE.")
            for ref in refs:
                ref.name += " [STALE]"
                ref.save()
            continue

        for ref in refs:
            for token in tokens[corp]:
                try:
                    esi_ref = client.Universe.get_universe_structures_structure_id(
                        structure_id=ref.structure_id,
                        token=token.valid_access_token()
                    ).results()
                    ref.name = esi_ref['name']
                    ref.save()
                    # Break the loop once we have successfully updated with a valid token
                    break
                except Exception as e:
                    logger.debug(f"Unable to get structure name with token: {token}")
                    logger.debug(e)
                    continue


@shared_task()
def update_observers():
    """
    Updates the observer status of all refineries.
    :return:
    """

    client = esi.client

    corps = Refinery.objects.all().values_list('corp__corporation_id', flat=True)
    # Build a dict of tokens to try for each corp.
    tokens = dict()
    for corp in corps:
        ts = _get_corp_tokens(corp, ESI_CHARACTER_SCOPES)
        if ts:
            tokens[corp] = ts

    for corp in tokens:
        observers = None
        # Get corp refineries
        refineries = Refinery.objects.filter(corp__corporation_id=corp)
        if len(tokens[corp]) == 0:
            # If we have no tokens for this corp, none of the refineries are valid observers
            for ref in refineries:
                ref.observer = False
                ref.save()
            continue

        for token in tokens[corp]:
            try:
                observers = client.Industry.get_corporation_corporation_id_mining_observers(
                    corporation_id=corp,
                    token=token.valid_access_token()
                ).results()
                # We can break the loop once we have a working token.
                break
            except Exception as e:
                logger.debug(f"Exception getting observer with token: {token}")
                logger.debug(e)
                continue
        if observers is not None:
            observer_ids = [observer['observer_id'] for observer in observers]
            for ref in refineries:
                # By default refineries are created assuming they are observers, so we only need to check
                # if they are missing from the list.
                if ref.structure_id not in observer_ids:
                    ref.observer = False
                    ref.save()


@shared_task()
def update_ledger():
    """
    Pulls mining ledger data from observers.
    :return:
    """
    client = esi.client

    corps = Refinery.objects.filter(observer=True).values_list('corp__corporation_id', flat=True)

    # Build a dict of tokens for each corp that can be tried.
    tokens = dict()
    for corp in corps:
        ts = _get_corp_tokens(corp, ESI_CHARACTER_SCOPES)
        if ts:
            tokens[corp] = ts

    for corp in tokens:
        if len(tokens[corp]) == 0:
            # This should never happen, but we should still make sure to handle it.
            continue
        # Get the observers for the current corp.
        observers = Refinery.objects.filter(observer=True, corp__corporation_id=corp)

        for observer in observers:
            # Reset ledger to an empty tuple on each iteration of the loop.
            ledger = ()
            for token in tokens[corp]:
                # Try to get the ledger with each token until one works.
                try:
                    ledger = client.Industry.get_corporation_corporation_id_mining_observers_observer_id(
                        corporation_id=corp,
                        observer_id=observer.structure_id,
                        token=token.valid_access_token()
                    ).results()
                    # Once a working token has been found/used we can break the token loop.
                    break
                except Exception as e:
                    # If no working token is found, we will just skip this observer. The next
                    # update refinery task should catch and handle this.
                    logger.debug(f"Exception getting ledger entries using token: {token}")
                    logger.debug(e)
                    continue

            if len(ledger) != 0:
                entries = []
                for entry in ledger:
                    # Change the type_id to be evetype
                    entry['evetype_id'] = entry.pop('type_id')
                    try:
                        LedgerEntry.objects.update_or_create(observer=observer, **entry)
                    except Exception as e:
                        logger.debug(f"Error creating entry: {entry}")
                        logger.debug(e)
                        continue
    update_active_extractions.delay()


@shared_task()
def update_active_extractions():
    """
    Updates flags for active extractions.
    :return:
    """
    # Get active extractions
    extractions = Extraction.objects.filter(active=True)

    # Loop over extractions
    for extraction in extractions:
        # First check if we need to set the jackpot flag.
        entries = LedgerEntry.objects.filter(
            observer=extraction.refinery,
            last_updated__lte=extraction.despawn,
            last_updated__gte=extraction.arrival_time,
            evetype__dogma_attributes__eve_dogma_attribute=2699,
            evetype__dogma_attributes__value=5,
        ).aggregate(jackpot=Count('id'))
        if entries['jackpot'] > 0:
            extraction.jackpot = True

        # Check if extraction is past despawn time (set not active)
        if datetime.datetime.utcnow().replace(tzinfo=pytz.utc) > extraction.despawn:
            extraction.active = False

        # Check if the extraction has been mined out (set not active)
        if extraction.total_volume is not None:
            entries = LedgerEntry.objects.filter(
                observer=extraction.refinery,
                last_updated__gte=extraction.arrival_time,
                last_updated__lte=extraction.despawn,
            )\
                .annotate(volume=Sum(F('quantity') * F('evetype__volume'), output_field=BigIntegerField()))\
                .aggregate(mined_volume=Sum('volume'))
            if entries['mined_volume'] >= extraction.total_volume:
                extraction.depleted = True
                extraction.active = False

        extraction.save()



from unittest import mock
from django.test import TestCase
from ..parser import ScanParser
from .parser_strings import with_header, without_header, with_quad_spaces, malformed


class TestScanParser(TestCase):
    def setUp(self):
        self.expected_result = {
            40217116: [
                {'ore_id': 45493, 'quantity': '0.239822223783', 'moon_id': 40217116},
                {'ore_id': 45499, 'quantity': '0.557235956192', 'moon_id': 40217116},
                {'ore_id': 45490, 'quantity': '0.202941834927', 'moon_id': 40217116}
            ]
        }

    def test_with_header(self):
        """
        Test the parser with scan data that contains a header.
        :return:
        """
        result = ScanParser(with_header).parse()
        self.assertEqual(result, self.expected_result)

    def test_without_header(self):
        """
        Test that the ScanParser will handle scan data that does not contain a header.
        :return:
        """
        result = ScanParser(without_header).parse()
        self.assertEqual(result, self.expected_result)

    def test_with_quad_spaces(self):
        """
        Test that the ScanParser will handle data that has 4 spaces rather than tabs.
        :return:
        """
        result = ScanParser(with_quad_spaces).parse()
        self.assertEqual(result, self.expected_result)

    def test_malformed(self):
        """
        Test that the ScanParser will raise an exception when the data is malformed.
        :return:
        """
        parser_obj = ScanParser(malformed)
        self.assertRaises(Exception, parser_obj.parse)

    def test_string_from_tabs(self):
        """
        Test that the __str__ method returns the original scan.
        :return:
        """
        scan_str = ScanParser(with_header).__str__()
        self.assertEqual(scan_str, with_header)

    def test_string_from_quad_spaces(self):
        """
        Test that the __str__ method returns the original scan but converted to tabbed.
        :return:
        """
        scan_str = ScanParser(with_quad_spaces).__str__()
        self.assertEqual(scan_str, with_header)



# Generated by Django 3.1.2 on 2021-04-20 06:20

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0014_auto_20210105_1413'),
        ('moonstuff', '0002_remove_material_model'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='trackingcharacter',
            name='id',
        ),
        migrations.AlterField(
            model_name='trackingcharacter',
            name='character',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='eveonline.evecharacter'),
        ),
    ]



from django.core.management.base import BaseCommand

from ...tasks import load_types_and_mats


class Command(BaseCommand):
    help = 'Preloads type and material data required for moonstuff.'

    def handle(self, *args, **options):
        categories = [25]
        groups = [18, 423, 427, 1406]

        self.stdout.write("Moonstuff Data Loader")
        self.stdout.write("=====================")

        self.stdout.write("This command will start loading type and material data required by the "
                          "moonstuff app.")

        self.stdout.write("This process can take a while to complete "
                          "and can create significant load on your system.")

        answer = input("Are you sure you would like to proceed? [y/N]: ")
        if answer.lower() == 'y':
            load_types_and_mats.delay(
                category_ids=categories, group_ids=groups, type_ids=None, force_loading_dogma=True
            )
            self.stdout.write(self.style.SUCCESS("Data loading has been started."))
        else:
            self.stdout.write(self.style.WARNING("Aborted."))



from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from . import urls


class MoonstuffMenu(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(self,
                              'Moon Tools',
                              'fas fa-moon',
                              'moonstuff:dashboard',
                              navactive=['moonstuff:'])

    def render(self, request):
        if request.user.has_perm('moonstuff.access_moonstuff'):
            return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_menu():
    return MoonstuffMenu()


@hooks.register('url_hook')
def register_url():
    return UrlHook(urls, 'moonstuff', '^moons/')



from datetime import datetime, timedelta

from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required, permission_required
from typing import Iterable
from django.contrib import messages
from django.utils.translation import gettext as gt
from django.conf import settings

from esi.decorators import token_required
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

from .tasks import process_scan, import_extraction_data
from .models import Resource, TrackingCharacter, Extraction, EveMoon
from .providers import ESI_CHARACTER_SCOPES

logger = get_extension_logger(__name__)

# Get refine setting
refine = .876
if hasattr(settings, 'MOON_REFINE_PERCENT'):
    if settings.MOON_REFINE_PERCENT > 1:
        refine = settings.MOON_REFINE_PERCENT / 100
    else:
        refine = settings.MOON_REFINE_PERCENT

# Get Default Extraction View setting
extraction_view = "Calendar"
if hasattr(settings, "DEFAULT_EXTRACTION_VIEW"):
    extraction_view = settings.DEFAULT_EXTRACTION_VIEW


def _get_resource_values(resources: Iterable[Resource]) -> dict:
    """
    Returns a dict containing the per-m3 values for a given list of resources.
    :param resources:
    :return:
    """
    ret = dict()

    for resource in resources:
        value = 0
        mats = resource.ore.materials.all()
        for mat in mats:
            ore_volume = resource.ore.volume
            amount = mat.quantity
            mat_value = mat.material_eve_type.market_price.average_price
            value += (((amount / 100) * refine) * mat_value) / ore_volume
        ret[resource.ore.id] = value

    return ret


def _get_moon_value_dict(moon_id: int) -> dict:
    """
    Returns a dict containing the per-m3 values of the moon's resources
    :param moon_id: The id of the moon.
    :return:
    """
    resources = Resource.objects\
        .prefetch_related('ore', 'ore__materials', 'ore__materials__material_eve_type__market_price')\
        .filter(moon__id=moon_id)

    ret = _get_resource_values(resources)

    return ret


def _get_extractions(limit=None):
    """
    Gets a dict of extractions from beginning of the current.
    :param limit: Number of days out to go. (Default: None - Will grab ALL extractions)
    :return:
    """
    if limit:
        qs = Extraction.objects.select_related('moon')\
            .filter(arrival_time__gte=datetime.utcnow().replace(day=1),
                    arrival_time__lte=datetime.utcnow()+timedelta(days=limit),
                    cancelled=False)\
            .prefetch_related('moon__resources', 'moon__resources__ore', 'refinery')\
            .order_by('arrival_time')
    else:
        qs = Extraction.objects.select_related('moon')\
            .filter(arrival_time__gte=datetime.utcnow().replace(day=1),
                    cancelled=False)\
            .prefetch_related('moon__resources', 'moon__resources__ore', 'refinery')\
            .order_by('arrival_time')

    return qs


def _build_event_dict(qs):
    ret = [
        {"title": q.refinery.name,
         "start": datetime.strftime(q.arrival_time, '%Y-%m-%dT%H:%M:%S%z'),
         "moon": q.moon.name,
         "rarity": [r.rarity for r in q.moon.resources.all()],
         "moon_id": q.moon.id}
        for q in qs
    ]

    return ret


# Create your views here.
@login_required
@permission_required('moonstuff.access_moonstuff')
def dashboard(request):
    """
    The main view for moonstuff.
    :param request: HTTPRequest object
    :return:
    """
    ctx = dict()

    # Get upcoming extraction events (calendar)
    extractions = _get_extractions()
    events = _build_event_dict(extractions)

    # Get moons
    moons = EveMoon.objects.filter(resources__isnull=False).distinct()\
        .prefetch_related('resources',
                          'resources__ore',
                          'resources__ore__materials',
                          'resources__ore__materials__material_eve_type__market_price',
                          'extractions',
                          'extractions__refinery',
                          'extractions__refinery__corp',
                          'eve_planet',
                          'eve_planet__eve_solar_system',
                          'eve_planet__eve_solar_system__eve_constellation__eve_region',
                          )

    resources = tuple(set(res for moon in moons for res in moon.resources.all()))
    ctx['events'] = events
    ctx['extractions'] = extractions
    ctx['moons'] = moons
    ctx['resources'] = _get_resource_values(resources)
    ctx['default_view'] = extraction_view
    return render(request, 'moonstuff/dashboard.html', ctx)


@login_required
@permission_required('moonstuff.add_resource')
def add_scan(request):
    """
    View for adding moon scan data.
    :param request: HTTPRequest object
    :return:
    """
    if request.method == 'POST':
        scan_data = request.POST['scan']

        process_scan.delay(scan_data, request.user.id)
        messages.success(request, gt('Your moon scan is being processed. Depending on size this may take some time.'))
        return redirect('moonstuff:dashboard')

    return render(request, 'moonstuff/add_scan.html')


@login_required
@token_required(scopes=ESI_CHARACTER_SCOPES)
@permission_required('moonstuff.add_trackingcharacter')
def add_character(request, token):
    """
    View for adding tracking character and corresponding token.
    :param request: HTTPRequest object
    :param token: django-esi Token object
    :return:
    """

    eve_char = EveCharacter.objects.get(character_id=token.character_id)
    if not TrackingCharacter.objects.filter(character=eve_char).exists():
        messages.success(request, gt('Character added!'))
        char = TrackingCharacter(character=eve_char)
        char.save()

        # Schedule an import task to pull data from the new Tracking Character.
        import_extraction_data.delay()
    else:
        messages.error(request, gt('That character is already being tracked!'))

    return redirect('moonstuff:dashboard')


@login_required
@permission_required('moonstuff.access_moonstuff')
def moon_info(request, moon_id=None):
    """
    View for viewing a moon's data.
    :param request: HTTPRequest object
    :param moon_id: integer
    :return:
    """
    ctx = {}
    if moon_id is None:
        messages.error(request, gt("Please provide the ID of a moon to view it's data."))
        return redirect('moonstuff:dashboard')

    # Get moon
    try:
        moon = EveMoon.objects.filter(id=moon_id, resources__isnull=False)\
            .prefetch_related('extractions',
                              'extractions__refinery',
                              'resources',
                              'resources__ore',
                              )[0]
        ctx['resources'] = _get_moon_value_dict(moon_id)
        ctx['moon'] = moon
    except EveMoon.DoesNotExist:
        messages.error(request, gt('A moon matching the provided ID could not be found.'))
        return redirect('moonstuff:dashboard')

    if request.headers.get('x-requested-with') == 'XMLHttpRequest':
        return render(request, 'moonstuff/moon_info_ajax.html', ctx)
    return render(request, 'moonstuff/moon_info.html', ctx)



class ScanParser:
    def __init__(self, scan_data: str):
        if "    " in scan_data:
            self.scan = self.__tabify(scan_data)
        else:
            self.scan = scan_data

    def __tabify(self, scan_data: str) -> str:
        """
        Converts quad-space separated data back to tab separated data.
        """
        return scan_data.replace("    ", "\t")

    def parse(self) -> dict:
        """
        Parses the scan into a dict that can then be processed.
        """
        ret = dict()

        # Split the scan into lines, ensure any CR characters are removed.
        scan_lines = self.scan.strip("\r").split("\n")

        # Remove any header lines (including moon names)
        scan_lines = list(line for line in scan_lines if "Moon" not in line)

        # Build dict of resources keyed to moon id
        for line in scan_lines:
            lst = line.split("\t")[1::]  # Trim off the leading blank space
            if len(lst) < 6:
                raise Exception("The moonscan provided appears to be malformed.")
            moon_id = int(lst[-1])
            ore_id = int(lst[2])
            quantity = lst[1]
            if moon_id not in ret.keys():
                ret[moon_id] = list()
            ret[moon_id].append({"ore_id": ore_id, 'quantity': quantity, 'moon_id': moon_id})

        return ret

    def __str__(self) -> str:
        return self.scan

    def __repr__(self) -> str:
        return repr(self.scan)
